/*****************************************************************************
 * A x e l
 *****************************************************************************
 * QArrangement
 * 2006-11-16
 * Julien Wintz
 *****************************************************************************
 *               Copyright (C) 2006 INRIA Sophia-Antipolis
 *****************************************************************************
 * Comments :
 ****************************************************************************/

/*! \brief A generic arrangement class implemented as a deamon for Axel.
    \class QArrangement QArrangement.h QArrangement.cpp
 
Generic arrangement class. We provide type definitions to fall back into
arrangement terminology but this is not really mandatory.

Pure abstract methods are object-dependant, which means, their 
implementation differs regarding the type of input objects.

Since the arrangement is dynamic, its data structure is an augmented 
influence graph. Therefore the arrangement can be maintained under 
additions or removals of objects.

The algorithm is implemented as a deamon (thred) that can be controlled.

*****************************************************************************/

# include <QString>
# include <QSet>

# include <QAxel/QCurve.h>
# include <QAxel/QSurface.h>

# include <QGui/color.h>

# include <QArrangement/QArrangement.h>

using namespace std ;

namespace QArrangement {

QArrangement::QArrangement(QString type, double epsilon, double epsilon_max) : c_epsilon(epsilon), c_epsilon_max(epsilon_max)
{
    m_stop = false ;
    m_pause = true ;
    a_type = type ;
    a_graph = new QAugmentedInfluenceGraph ;
    a_policy = Center ;
    
    a_root = new QGraphNode("") ;
    a_root->setColors(Qt::magenta, Qt::darkMagenta) ;
    a_graph->addNode(a_root) ;
    
    gverbose_cells   = false ;
    gverbose_fusion  = false ;
}

QArrangement::~QArrangement(void)
{
	
}
	
void QArrangement::play(void) 
{
    m_pause = false ; m_pauseCondition.wakeAll() ;
}

void QArrangement::pause(void)
{
    m_pause = true ;
}

void QArrangement::stop(void)
{
    m_stop = true ;
    m_pause = false ;
}

void QArrangement::setTree(QAxelObject * o)
{
    if(a_trees[o])
	emit display(a_trees[o]) ;
}

QGraph * QArrangement::graph(void) 
{ 
    return a_graph ; 
}

void QArrangement::init(QList<Object_t *> objects)
{
    foreach(Object_t * o, objects)
	addObject(o) ;
}

void QArrangement::run(void)
{
    // qDebug("Thread start...") ;
    // 
    // bool r_init = false ;
    // 
    // forever {
    // 	if(!r_init) {
    init(mainwindow->viewer()->listObjectsFromType(a_type)) ;
    // 		r_init = true ;
    // 	}
    // 	if (m_stop)  break ;
    // 	if (m_pause) m_pauseCondition.wait(&m_mutex) ;
    //   
    // 	sleep(1) ;
    // 	qDebug("Arrangement Algorithm running...") ;
    // }
    // 
    // qDebug("Thread stop...") ;
}

void QArrangement::addObject(QAxelObject * o) 
{
    if(o->type() != a_type) return ;

    QList<Region_t *> regions = getRegions(o) ;
    
    foreach(Region_t * r, regions) {
	QGraphNode * n = a_graph->insert(r, a_root) ;

 	QList<GraphNode_t *> killed = locate(r) ;

 	foreach(GraphNode_t * gn, killed) {
//	    gn->setColors(Qt::magenta, Qt::darkMagenta) ;
	    
	    if(gn != a_root) {

//     cout << FG_YELLOW << "Region " << r->name().toAscii().constData() << " " << "determined by" << flush ;
//     foreach(QAxelObject * o, r->determinors()) {
// 	cout << " " <<  o->name().toAscii().constData() << flush ;	    
//     }
//     cout << NOCOLOR << endl ;
		
//     cout << FG_MAGENTA << "Region " << gn->region()->name().toAscii().constData() << " " << "determined by" << flush ;
//     foreach(QAxelObject * o, gn->region()->determinors()) {
// 	cout << " " <<  o->name().toAscii().constData() << flush ;	    
//     }
//     cout << NOCOLOR << endl ;
		
		QList<Region_t *> intersections = r->intersect(gn->region()) ;
		foreach( Region_t * intersection, intersections) {
		    a_graph->insert(intersection, n, gn) ;
		}
	    }
	}
    }
}

void QArrangement::removeObject(Object_t * o) 
{
    Q_UNUSED(o) ;
}

QList<Region_t *> QArrangement::getRegions(Object_t * o)
{
    QList<Node_t *> nodes ;
    Octree_t a_octree ;
    Node_t * root = a_octree.GetRoot() ;
    Cell_t * cell = NULL ;
    if(a_type.contains("Curve"))   cell = new QCell(((QCurve *)o)->boundingbox()) ;
    if(a_type.contains("Surface")) cell = new QCell(((QSurface *)o)->boundingbox()) ;
    root->SetCell(cell) ;	
    nodes.push_back(root) ;
    // mainwindow->viewer()->addObject(cell) ;
    
    while(!nodes.empty()) {
	Node_t * no = nodes.takeFirst() ;
	Cell_t * bb = no->cell ;
	if(gverbose_cells) {
	    bb->blacklist() ;
	    mainwindow->viewer()->addObject(bb) ;
	}		    
	if(bb->size() < c_epsilon_max && regular(o, no)) {
	    no->objects << topology(o, no) ;
	} else if(bb->size() > c_epsilon) {
	    QList<Cell_t *> cells = subdivide(o, no) ;

	    // The neighborhood is already set for the nodes, so, each functions using neighborhood relations have to use a node instead of a cell
	    no->F_NWchild = new Node_t(F_NW, no, cells[0]) ;
	    no->F_NEchild = new Node_t(F_NE, no, cells[1]) ;
	    no->F_SWchild = new Node_t(F_SW, no, cells[2]) ;
	    no->F_SEchild = new Node_t(F_SE, no, cells[3]) ;
	    
	    nodes << no->F_NWchild << no->F_NEchild << no->F_SWchild << no->F_SEchild ;
	} else {
	    qCritical("Beyond minimum subdivision size, generating dummy region") ;
	    no->objects << dummy(no) ;
	}
    }
    
    a_trees.insert(o, a_octree.toGraph()) ;
    emit display(a_trees.value(o)) ;

    QList<Region_t *> regions = fusion(a_octree) ;
    foreach(Region_t * r, regions) {
	r->setCell(cell) ; 
	r->addDeterminor(o) ;
    }
    return regions ;
}

QList<Region_t *> QArrangement::fusion(Octree_t octree)
{
    return fusion(octree.GetRoot()) ;
}

QList<Region_t *> QArrangement::fusion(Node_t * node)
{
    if(node->isLeaf()) return fusion(node->objects) ;
    else return fusion(QList<Region_t *>()
		       << fusion(QList<Region_t *>() 
				 << fusion(node->GetF_NWchild()) 
				 << fusion(node->GetF_NEchild()))
		       << fusion(QList<Region_t *>() 
				 << fusion(node->GetF_SWchild()) 
				 << fusion(node->GetF_SEchild()))
	) ;
}

QList<Region_t *> QArrangement::fusion(QList<Region_t *> & regions)
{
    int i = 0 ;
    while(i < regions.size()) {
	int j = i+1 ;
	while( j < regions.size()) {
	    if(i != j) {
		if(gverbose_fusion) {
		    regions[i]->setColor(Qt::yellow) ;
		    mainwindow->viewer()->addObject(regions[i]) ;
		    regions[j]->setColor(Qt::blue) ;
		    mainwindow->viewer()->addObject(regions[j]) ;
		    getchar() ;
		    mainwindow->viewer()->removeObject(regions[i]) ;
		    mainwindow->viewer()->removeObject(regions[j]) ;
		}
	    }
	    
	    if(i != j && regions[i]->mergesWith(regions[j], a_type)) {
		regions[i]->mergeWith(regions[j]) ;
		if(gverbose_fusion) {
		    regions[i]->setColor(Qt::green) ;
		    mainwindow->viewer()->addObject(regions[i]) ;
		    getchar() ;
		    mainwindow->viewer()->removeObject(regions[i]) ;
		}
		regions.removeAt(j) ;
		j = i+1 ;
	    } else j++ ;
	}
	i++ ;
    }
    return regions ;
}

Vertex_t * QArrangement::xsmallest(QList<Vertex_t *>& l) 
{
    Vertex_t * s = l[0] ; int n = 0, i = 0 ;
    foreach(Vertex_t * p, l) {
	if(p->x() < s->x()) {
	    s = p ; n = i ;
	}
	i++ ;
    }
    return l.takeAt(n) ;
}

Vertex_t * QArrangement::xgreatest(QList<Vertex_t *>& l) 
{
    Vertex_t * s = l[0] ; int n = 0, i = 0 ;
    foreach(Vertex_t * p, l) {
	if(p->x() > s->x()) {
	    s = p ; n = i ;
	}
	i++ ;
    }
    return l.takeAt(n) ;
}

Vertex_t * QArrangement::ysmallest(QList<Vertex_t *>& l) 
{
    Vertex_t * s = l[0] ; int n = 0, i = 0 ;
    foreach(Vertex_t * p, l) {
	if(p->y() < s->y()) {
	    s = p ; n = i ;
	}
	i++ ;
    }
    return l.takeAt(n) ;
}

Vertex_t * QArrangement::ygreatest(QList<Vertex_t *>& l) 
{
    Vertex_t * s = l[0] ; int n = 0, i = 0 ;
    foreach(Vertex_t * p, l) {
	if(p->y() > s->y()) {
	    s = p ; n = i ;
	}
	i++ ;
    }
    return l.takeAt(n) ;
}

bool QArrangement::isCellBorder(Vertex_t * p, Cell_t * bb)
{
    if(p->x() == bb->xmin() && p->y() == bb->ymin()) return true ;
    if(p->x() == bb->xmin() && p->y() == bb->ymax()) return true ;
    if(p->x() == bb->xmax() && p->y() == bb->ymin()) return true ;
    if(p->x() == bb->xmax() && p->y() == bb->ymax()) return true ;
    return false ;
}

QList<Vertex_t *> QArrangement::clockwiseSubdivision(QList<Point_t *> & intersectionPoints, Cell_t * bb, Vertex_t * nw, Vertex_t * ne, Vertex_t * se, Vertex_t * sw)
{
    QList<Vertex_t *> northList ;
    northList << nw ;
    foreach(Point_t * ipt, intersectionPoints)
	if(ipt->y() == bb->ymax()) {
	    northList << new Vertex_t(ipt) ;
	    intersectionPoints.removeAll(ipt) ;
	}
    northList << ne ;
    
    QList<Vertex_t *> eastList ;
    eastList << ne ;
    foreach(Point_t * ipt, intersectionPoints)
	if(ipt->x() == bb->xmax()) {
	    eastList << new Vertex_t(ipt) ;
	    intersectionPoints.removeAll(ipt) ;
	}
    eastList << se ;
    
    QList<Vertex_t *> southList ;
    southList << se ;
    foreach(Point_t * ipt, intersectionPoints)
	if(ipt->y() == bb->ymin()) {
	    southList << new Vertex_t(ipt) ;
	    intersectionPoints.removeAll(ipt) ;
	}
    southList << sw ;
    
    QList<Vertex_t *> westList ;
    westList << sw ;
    foreach(Point_t * ipt, intersectionPoints)
	if(ipt->x() == bb->xmin()) {
	    westList << new Vertex_t(ipt) ;
	    intersectionPoints.removeAll(ipt) ;
	}
    westList << nw ;
    
    northList.removeFirst() ;
    eastList.removeFirst() ;
    southList.removeFirst() ;
    westList.removeFirst() ;
    
    QList<Vertex_t *> rvertices ;
    while(!northList.empty()) rvertices << xsmallest(northList) ;
    while(!eastList.empty()) rvertices << ygreatest(eastList) ;
    while(!southList.empty()) rvertices << xgreatest(southList) ;
    while(!westList.empty()) rvertices << ysmallest(westList) ;
    
    // Ensure rvertices does not begin with a cell border ...
    while(isCellBorder(rvertices.first(), bb)) rvertices << rvertices.takeFirst() ;
    
    // Append the first at the beginning ...
    rvertices << rvertices.first() ;
    
    return rvertices ;
}

Region_t * QArrangement::dummy(Node_t * n)
{
    QCell * bb = n->GetCell() ;
    
    Vertex_t * se = new Vertex_t(bb->xmax(), bb->ymin()) ; se->setName("se") ;
    Vertex_t * sw = new Vertex_t(bb->xmin(), bb->ymin()) ; sw->setName("sw") ;
    Vertex_t * ne = new Vertex_t(bb->xmax(), bb->ymax()) ; ne->setName("ne") ;
    Vertex_t * nw = new Vertex_t(bb->xmin(), bb->ymax()) ; nw->setName("nw") ;
    
    Region_t * f = new Region_t() ;
    f->addVertices(QList<Vertex_t *>() << nw << sw << se << ne) ;
    f->addEdge(nw, sw, bb) ;
    f->addEdge(sw, se, bb) ;
    f->addEdge(se, ne, bb) ;
    f->addEdge(ne, nw, bb) ;
    
    return f ;
}

QList<GraphNode_t *> QArrangement::locate(Point_t * p)
{
    return locate(p, a_root) ;
}

QList<GraphNode_t *> QArrangement::locate(Point_t * p, GraphNode_t * n)
{			
    QList<GraphNode_t *> l ;
    
    if(n == a_root) {
	if(n->isLeaf())
	    l << n ;
	else {
	    QList<GraphNode_t *> t ;
	    foreach(GraphNode_t * c, n->children())
		t << locate(p, c) ;
	    if(t.empty())
		l << n ;
	    else
		l << t ;
	}
	return l ;
    }

    if(n->region()->contains(p)) {
	if(n->isLeaf()) 
	    l << n ;
	else {
	    QList<GraphNode_t *> t ;
	    foreach(GraphNode_t * c, n->children()) 
		t << locate(p, c) ;
	    if(t.empty()) 
		l << n ;
	    else 
		l << t ;
	}
    }

    QColor orange(255, 69, 0) ;
    QColor darkOrange(205, 55, 0) ;

    foreach(GraphNode_t * n, l)
	n->setColors(orange, darkOrange) ;

    a_graph->update() ;

    return l ;
}

QList<GraphNode_t *> QArrangement::locate(Region_t * r)
{
    return locate(r, a_root) ;
}

QList<GraphNode_t *> QArrangement::locate(Region_t * r, GraphNode_t * n)
{			
    QList<GraphNode_t *> l ;
    
    /* If we are visiting the root */
    if(n == a_root) {
	if(n->isLeaf()) 
	    l << n ;
	else {
	    QList<GraphNode_t *> t ;
	    foreach(GraphNode_t * c, n->children()) 
		t << locate(r, c) ;
	    if(t.empty()) 
		l << n ;
	    else 
		l << t ;
	}
	return l ;
    }

    /* The region has been determined by the same region in the node */
    if(r->determinors().size() != n->region()->determinors().size()) 
	return l ;
    else {
	bool sameDeterminors = true ;

	foreach(QAxelObject * o, r->determinors()) {
	    if(!n->region()->determinors().contains(o))
		sameDeterminors = false ;
	}

	if(sameDeterminors) return l ;
    }

    /* Here we really check for conflicts */
    if(n->region()->intersects(r)) {
	if(n->isLeaf()) 
	    l << n ;
	else {
	    QList<GraphNode_t *> t ;
	    foreach(GraphNode_t * c, n->children()) 
		t << locate(r, c) ;
	    if(t.empty()) 
		l << n ;
	    else 
		l << t ;
	}
    }
    return l ;
}

} ; // namespace QArrangement ;
