/*****************************************************************************
 * A x e l
 *****************************************************************************
 * QMaterial
 * 2007-06-17
 * Julien Wintz
 *****************************************************************************
 *               Copyright (C) 2006 INRIA Sophia-Antipolis
 *****************************************************************************/

# ifndef QMATERIAL_H
# define QMATERIAL_H

# include <QAxel/QAxelObject.h>

namespace QAxel {

/****************************************************************************/
/*!
  \class QMaterial QMaterial.h
  \brief This class defines an OpenGL material.

  This class is used to encaspulate all functionality needed for a material
  in \e OpenGL. A material contains the following properties :

  <ul>
  <li>a \e diffuse color. This is the main color of the material.
  <li>a \e specular color. Used for the highlights.
  <li>a \e ambient color. This is a color considered to 'always be around'.
  <li>a \e emissive color. The material is considered to be emitting this
      color to the environment.
  <li>\e shininess. The shininess is basically how 'polished' the material is.
  </ul>

  A method exists for rendering the material. This requires that a OpenGL-
  rendering context is valid. Other method lets the user access the 
  individual components of the material, as well as an overloaded
  \e =-operator.
*/
/****************************************************************************/
class QMaterial : public QAxelObject 
{
    Q_OBJECT

public:

    QMaterial(QAxelObject * parent = 0) ;
    QMaterial(QAxelObject * parent,
	      const QColor & ambient, 
	      const QColor & diffuse,
	      const QColor & specular,
	      const QColor & emissive,
	      const float Shininess) ;
    QMaterial(QAxelObject * parent,
	      const QMaterial & m) ;

    void draw(void) ;
    void drawWithNames(void) ;

    const QColor & ambient(void) const ;
    const QColor & diffuse(void) const ;
    const QColor & specular(void) const ;
    const QColor & emissive(void) const ;
    float shininess(void) const ;

    void setAmbient(const QColor & ambient) ;
    void setDiffuse(const QColor & diffuse) ;
    void setSpecular(const QColor & specular) ;
    void setEmissive(const QColor & emissive) ;
    void setShininess(const float shininess) ;

    QMaterial & operator = (const QMaterial & m) ;

private:

  QColor m_Ambient ;
  QColor m_Diffuse ;
  QColor m_Specular ;
  QColor m_Emissive ;

  float m_Shininess ;
} ;

} ; // namespace QAxel

# endif  // QMATERIAL_H
 
