/******************************************************************************
 * A x e l                                                                   *
 ******************************************************************************
 * QRegion3f                                                                 *
 * 2006-10-14                                                                *
 * Julien Wintz                                                              *
 ******************************************************************************
 *               Copyright (C) 2006 INRIA Sophia-Antipolis                   *
 ******************************************************************************
 * Comments :                                                                *
 *****************************************************************************/

// # define INTDEBUG 1
// # define CPPDEBUG 1
// # define NRGDEBUG 1

# include <QMap>
# include <QHash>

# include <QGui/color.h>
# include <QGui/Viewer.h>

# include <QData/QCell.h>
# include <QData/QRegion3f.h>
# include <QData/QRegionVertex.h>
# include <QData/QRegionEdge.h>
# include <QData/QRegionFace.h>
# include <QData/QRegionSegmentation.h>

# include <QAxel/QCylinder.h>
# include <QAxel/QCurve.h>
# include <QAxel/QArrow.h>
# include <QAxel/ObjectManager.h>
# include <QAxel/PluginManager.h>

# include <QGLViewer/vec.h>

# include <QSynaps/QSynapsInterface.h>

# include <QCgal/QCgalInterface.h>

using namespace std ;

namespace QData {

int QRegion3f::id_ = 0 ;

QRegion3f::QRegion3f(void)
{
    this->r_segmentation = new QRegionSegmentation(this) ;
    this->r_dlist        = new QDisplayList ; 
    this->r_dirty        = false ;

    QString str ;
    setProperty("type", "Region3f") ;
    setProperty("name", "f_"+str.setNum(id_)) ;
    id_++ ;
    setObjectColor(new QColor(255,55,98)) ;
}

QRegion3f::~QRegion3f(void)
{
	
}

void QRegion3f::draw()
{
    QCgalInterface * iCgal = dynamic_cast<QCgalInterface *>(PluginManager::instance()->plugin("Cgal")) ;
    Q_UNUSED(iCgal) ;

    if(r_dirty || !r_dlist->isValid()) {
	r_dlist->begin() ;
	foreach(QRegionEdge * e, r_edges) {
	    QArrow * arrow = new QArrow(e->source(), e->dest(), 0.005) ;
	    arrow->setColor(*color()) ;
	    arrow->draw() ;
	}
	
	foreach(QRegionVertex * v, r_vertices) {
	    v->setSize(0.005) ;
	    v->draw() ;
	}
	
//	if(iCgal) iCgal->mesh(this)->draw() ;

	r_dlist->end() ;
    }
    r_dlist->draw() ;
}

void QRegion3f::drawWithNames()
{
	
}

QList<QRegionVertex *> QRegion3f::vertices(void) 
{
    return r_vertices ;
}

QList<QRegionEdge *>  QRegion3f::edges(void) 
{
    return r_edges ;
}

QList<QAxelObject *> QRegion3f::determinors(void)
{
    return r_determinors ;
}

void QRegion3f::addVertex(QRegionVertex * p)
{
    if(!r_vertices.contains(p))
	r_vertices << p ;
    r_dirty = true ;
}

void QRegion3f::addVertices(QList<QRegionVertex *> l) 
{
    foreach(QRegionVertex * p, l) 
	addVertex(p) ;
    r_dirty = true ;
}

void QRegion3f::removeVertex(QRegionVertex * v) 
{
    r_vertices.removeAll(v) ;
    r_dirty = true ;
}

void QRegion3f::addEdge(QRegionVertex * p1, QRegionVertex * p2, QAxelObject * o)
{
    QRegionEdge * e = new QRegionEdge(p1, p2, o) ;
    p1->addEdge(e) ;
    p2->addEdge(e) ;
    addVertex(p1) ;
    addVertex(p2) ;
    r_edges << e ;
    r_dirty = true ;
}

void QRegion3f::addEdge(QRegionEdge * e)
{
    if(!r_edges.contains(e))
	r_edges << e ;
    addVertex(e->source()) ;
    addVertex(e->dest()) ;
    r_dirty = true ;
}

void QRegion3f::addDeterminor(QAxelObject * o) 
{
    if(!r_determinors.contains(o))
	r_determinors << o ;
    r_dirty = true ;
}

void QRegion3f::addConflictor(QAxelObject * o)
{
    if(!r_conflictors.contains(o))
	r_conflictors << o ;
    r_dirty = true ;
}

QList<QRegionVertex *> QRegion3f::commonVerticesWith(QRegion3f * r)
{
    QList<QRegionVertex *> commons ;
	
    foreach(QRegionVertex * p, r_vertices)
	foreach(QRegionVertex * v, r->vertices())
	if(p->equals(v))     
	    commons << p ;
	       
    return commons ;
}

QList<QRegionVertex *> QRegion3f::commonVerticesOf(QRegion3f * r)
{
    QList<QRegionVertex *> commons ;
    
    foreach(QRegionVertex * p, r_vertices)
	foreach(QRegionVertex * v, r->vertices())
	if(p->equals(v))
	    commons << v ;
    
    return commons ;
}

QList<QRegionVertex *> QRegion3f::uncommonVerticesOf(QRegion3f * r)
{
    QList<QRegionVertex *> uncommons ;
    
    foreach(QRegionVertex * v, r->vertices()) {
	bool add = true ;
	foreach(QRegionVertex * p, r_vertices)
	    if(p->equals(v))
		add = false ;
	if(add) uncommons << v ;
    }
    
    return uncommons ;
}

QList<QRegionEdge *> QRegion3f::commonEdgesOf(QRegion3f * r)
{
    QList<QRegionEdge *> commons ;
    
    foreach(QRegionEdge * e1, r_edges)
	foreach(QRegionEdge * e2, r->edges()) {
	// cout << __func__ << " : [" << e1->source()->x() << "," << e1->source()->y() << "],[" << e1->dest()->x() << "," << e1->dest()->y() << "]"
	// 	 << " ???? [" << e2->source()->x() << "," << e2->source()->y() << "],[" << e2->dest()->x() << "," << e2->dest()->y() << "]"
	// 	 << flush ;
	
	if(e1->equals(e2)) {
	    commons << e2 ;
	    // cout << FG_GREEN << "Ok" << NOCOLOR << endl ;
	} else {
	    // cout << FG_RED << "No" << NOCOLOR << endl ;
	}
    }
    
    return commons ;
}

QList<QRegionEdge *> QRegion3f::commonEdgesWith(QRegion3f * r) 
{
    QList<QRegionEdge *> commons ;
    
    foreach(QRegionEdge * e1, r_edges) 
	foreach(QRegionEdge * e2, r->edges())
	if(e1->equals(e2))
	    commons << e1 ;
    
    return commons ;
}

QList<QRegionEdge *> QRegion3f::uncommonEdgesOf(QRegion3f * r)
{
    QList<QRegionEdge *> uncommons ;
    
    foreach(QRegionEdge * e2, r->edges()) {
	bool add = true ;
	foreach(QRegionEdge * e1, r_edges)
	    if(e1->equals(e2))
		add = false ;
	if(add) uncommons << e2 ;
    }
    
    return uncommons ;
}

/*!************************************************************************************
 * Returns the top most vertex defining the region, 
 * i.e. the vertex with max y coordinates.
 **************************************************************************************/
QRegionVertex * QRegion3f::topMostVertex(void)
{
    if(r_vertices.empty()) return NULL ;

    QRegionVertex * r = r_vertices.first() ;
    foreach(QRegionVertex * v, r_vertices)
	if(v->y()>r->y())
	    r = v ;
    return r ;
}

/*!************************************************************************************
 * Returns the bottom most vertex defining the region, 
 * i.e. the vertex with min y coordinates.
 **************************************************************************************/
QRegionVertex * QRegion3f::bottomMostVertex(void)
{
    if(r_vertices.empty()) return NULL ;

    QRegionVertex * r = r_vertices.first() ;
    foreach(QRegionVertex * v, r_vertices)
	if(v->y()<r->y())
	    r = v ;
    return r ;
}

/*!************************************************************************************
 * Returns the left most vertex defining the region, 
 * i.e. the vertex with min x coordinates.
 **************************************************************************************/
QRegionVertex * QRegion3f::leftMostVertex(void)
{
    if(r_vertices.empty()) return NULL ;

    QRegionVertex * r = r_vertices.first() ;
    foreach(QRegionVertex * v, r_vertices)
	if(v->x()<r->x())
	    r = v ;
    return r ;
}

/*!************************************************************************************
 * Returns the right most vertex defining the region, 
 * i.e. the vertex with max x coordinates.
 **************************************************************************************/
QRegionVertex * QRegion3f::rightMostVertex(void)
{
    if(r_vertices.empty()) return NULL ;

    QRegionVertex * r = r_vertices.first() ;
    foreach(QRegionVertex * v, r_vertices)
	if(v->x()>r->x())
	    r = v ;
    return r ;
}

/*!************************************************************************************
 * Returns all the edges containing the vertex \param v.
 **************************************************************************************/
QList<QRegionEdge *> QRegion3f::edgesWith(QRegionVertex * v)
{
    QList<QRegionEdge *> l ;
    foreach(QRegionEdge * e, r_edges)
	if(e->contains(v))
	    l << e ;
    return l ;
}

/*!************************************************************************************
 * Reorders the list of edges and vertices, in place.
 **************************************************************************************/
void QRegion3f::sort(void)
{
    if(!r_dirty) return ;

    for(int i = 0 ; i < r_edges.size()-2 ; i++)
	if(!r_edges[i]->dest()->equals(r_edges[i+1]->source()))
	    for(int j = i+2 ; j < r_edges.size() ; j++)
		if(r_edges[i]->dest()->equals(r_edges[j]->source())) {
		    r_edges.swap(i+1, j) ;
		    break ;
		}
    
    // Sort vertices
    QList<QRegionVertex *> vertices ;
    foreach(QRegionEdge * e, r_edges) vertices << e->source() ;
    r_vertices = vertices ;
    
    r_dirty = true ;

//     cout << FG_GREEN ;
//     cout << "r_edges.size() = " << r_edges.size() << endl ;
//     cout << "r_verti.size() = " << r_vertices.size() << endl ;
//     foreach(QRegionEdge * e, r_edges) cout << e << endl ;
//     foreach(QRegionVertex * e, r_vertices) cout << e << endl ;
//     cout << NOCOLOR << endl ;
}

/*!************************************************************************************
 * Returns the last index in the loop of vertex \param index.
 **************************************************************************************/
int QRegion3f::lastIndexInLoopFrom(int index) 
{
    if(r_dirty) sort() ;

    int i = index ;
    while(i < r_edges.size()-1)
	if(r_edges[i]->dest()->equals(r_edges[i+1]->source())) 
	    i++ ;
	else
	    break ;

    return i ;
}

/*!************************************************************************************
 * Returns the first index in the loop of vertex \param index.
 **************************************************************************************/
int QRegion3f::firstIndexInLoopFrom(int index) 
{
    if(r_dirty) sort() ;
 
    int i = index ;
    while(i > 0)
	if(r_edges[i-1]->dest()->equals(r_edges[i]->source())) 
	    i-- ;
	else
	    break ;
    
    return i ;
}

/*!************************************************************************************
 * Test whether a region contains holes
 **************************************************************************************/
bool QRegion3f::containsHoles(void)
{
    return (lastIndexInLoopFrom(0) < r_edges.size()-1) ;
}

/*!************************************************************************************
 * Test whether two regions can be merged
 **************************************************************************************/
bool QRegion3f::mergesWith(QRegion3f * r, QString type)
{
    Q_UNUSED(type) ;
    QList<QRegionEdge *> commons = commonEdgesOf(r) ;
    
    if(commons.size() == 0) return false ;
    
    foreach(QRegionEdge * e, commons)
	if(e->object()->type() == "ImplicitCurve2D")
	    return false ;

    return true ;
}

/*!************************************************************************************
 * Merging the region (r) with another (f)
 * 
 * - Get the double vertices of r and f
 * - create a map of this vertices from r -> f
 * - update the edges of f accordingly
 *	- in the edges of f
 *  - in the region f
 * - remove common edges in both r and f
 * - Append the remaining edges and vertices of in r
 **************************************************************************************/
void QRegion3f::mergeWith(QRegion3f * f)
{
    QList<QRegionVertex *> from = commonVerticesWith(f) ;
    QList<QRegionVertex *> to   = commonVerticesOf(f) ;
    
    QMap<QRegionVertex *, QRegionVertex *> map ;
    foreach(QRegionVertex * pf, from)
	foreach(QRegionVertex * pt, to)
	if(pf->equals(pt))
	    map.insert(pt, pf) ;
    
    f->applyMap(map) ;

    QList<QRegionVertex *> dummyVertices ;
    foreach(QRegionEdge * e, commonEdgesWith(f))
	dummyVertices << e->source() << e->dest() ;
    
//    deleteEdgesIn(map.keys()) ;
//    f->deleteEdgesIn(map.keys()) ;

    deleteEdgesIn(dummyVertices) ;
    f->deleteEdgesIn(dummyVertices) ;

//    deleteEdgesIn(commonEdgesWith(f)) ;
//    f->deleteEdgesIn(commonEdgesOf(f)) ;
    
    foreach(QRegionVertex * p, f->vertices())
	if(!from.contains(p))
	    r_vertices << p ;
    
    foreach(QRegionEdge * e, f->edges())
	r_edges << e ;
    
    deleteVerticesIn(fusionEdges()) ;

    r_dirty = true ;
}

/*!************************************************************************************
 * Fusion edges of the region
 **************************************************************************************/
QList<QRegionVertex *> QRegion3f::fusionEdges(void) 
{
    QList<QRegionVertex *> l ;
    int i = 0 ;
    while(i < r_edges.size()) {
	int j = i+1 ;
	while( j < r_edges.size()) {
	    if(i != j && r_edges[i]->mergesWith(r_edges[j])) {
		l << r_edges[i]->mergeWith(r_edges[j]) ;
		r_edges.removeAt(j) ;
		j = i+1 ;
	    } else j++ ;
	}
	i++ ;
    }
    
    r_dirty = true ;

    return l ;
}

void QRegion3f::applyMap(QMap<QRegionVertex*, QRegionVertex*> map)
{
    foreach(QRegionVertex * v1, r_vertices)
	foreach(QRegionVertex * v2, map.keys())
	if(v1->equals(v2)) {
	    foreach(QRegionEdge * e, r_edges) e->replace(v1, map.value(v2)) ;
	    removeVertex(v1) ;
	}

    r_dirty = true ;
}

bool isIn(QList<QRegionVertex *> l, QRegionVertex * p) 
{
    foreach(QRegionVertex* t, l)
	if(t->equals(p)) 
	    return true ;
    return false ;
}

bool isIn(QList<QRegionEdge *> l, QRegionEdge * e)
{
    foreach(QRegionEdge * t, l)
	if(t->equals(e))
	    return true ;
    return false ;
}

void QRegion3f::deleteVerticesIn(QList<QRegionVertex *> l)
{
    int i = 0 ;
    while(i < r_vertices.size())
	if(isIn(l, r_vertices[i]))
	    r_vertices.removeAt(i) ;
	else i++ ;
    
    r_dirty = true ;
}

void QRegion3f::deleteEdgesIn(QList<QRegionVertex *> l)
{	
    int i = 0 ;
    while(i < r_edges.size()) 
	if(isIn(l, r_edges[i]->source()) && isIn(l, r_edges[i]->dest()))
	    r_edges.removeAt(i) ;
	else i++ ;

    r_dirty = true ;
}

void QRegion3f::deleteEdgesIn(QList<QRegionEdge *> l)
{	
    int i = 0 ;
    while(i < r_edges.size()) 
	if(isIn(l, r_edges[i]))
	    r_edges.removeAt(i) ;
	else i++ ;
    
    r_dirty = true ;
}

QString QRegion3f::toString(void)
{
    QString text = ":: --- Region " + name() + " --- ::\n" ;
    text += ":: Vertices : \n" ;
    foreach(QRegionVertex * p, r_vertices) text += ":: " + p->toString() + "\n" ;
    text += ":: Edges : \n" ;
    foreach(QRegionEdge * e, r_edges) text += ":: " + e->toString() + "\n" ;
    text += "::------------ ::\n" ;
    return text ;
}

std::ostream& operator << (std::ostream& stream, QRegion3f & r) 
{ 
    return stream << r.toString().toLatin1().data() ; 
}

std::ostream& operator << (std::ostream& stream, QRegion3f * r) 
{ 
    return stream << r->toString().toLatin1().data() ; 
}

QCell * QRegion3f::cell(void) 
{ 
    return r_cell ; 
}

void QRegion3f::setCell(QCell * cell) 
{ 
    r_cell = cell ; 
}

QList<QRegion3f *> QRegion3f::intersect(QRegion3f * other)
{
//     cout << FG_BLUE  << this  << NOCOLOR << endl ;
//     cout << FG_GREEN << other << NOCOLOR << endl ;

# ifdef INTDEBUG
    cout << FG_YELLOW << "== Intersection of " << name().toAscii().constData() << " and " << other->name().toAscii().constData() << " ===" << endl ;
# endif

    QList<QRegion3f *> l ;

    QSynapsInterface * iSynaps = dynamic_cast<QSynapsInterface *>(PluginManager::instance()->plugin("Synaps")) ;

    QRegion3f * r1 = this ;
    QRegion3f * r2 = other ;

    QRegionSegmentation * r_seg1 =  this->segmentation() ; if(!r_seg1->isValid()) r_seg1->build() ;
    QRegionSegmentation * r_seg2 = other->segmentation() ; if(!r_seg2->isValid()) r_seg2->build() ;

    QList<QSegmentationQuery_t> query = QRegionSegmentation::query(r_seg1, r_seg2) ;

    QMap<float, QPoint3f *> cps ; // A map to fill consecutive point positions of the region "this", say r1
    QMap<float, QPoint3f *> cqs ; // A map to fill consecutive point positions of the region "other", say r2
    QHash<QPoint3f *, int > nrg ; // A map to know when we jump from a region to another
    QHash<QPoint3f *, QRegionVertex *> rvt ; // A map to store new intersection region vertices
    QHash<QPoint3f *, QAxelObject *> ops  ; // A map to store objects adjacent to intersection vertices in r1
    QHash<QPoint3f *, QAxelObject *> oqs  ; // A map to store objects adjacent to intersection vertices in r2

//    foreach(QSegmentationQuery_t q, query) {
    while(!query.empty()) {

	QSegmentationQuery_t q = query.takeFirst() ;

	QSegmentationNode_t * n1 = q.first ;
	QSegmentationNode_t * n2 = q.second ;
	
	QBoundingBox * bb = QBoundingBox::intersect(n1->cell, n2->cell) ; 
	QCell * cell = new QCell(bb) ;

	QRegionEdge * e1 = n1->object() ;
	QRegionEdge * e2 = n2->object() ;

	QList<QPoint3f *> intersection   = iSynaps->intersection(e1, e2, bb) ;     

	if(intersection.size() == 0) {

	    continue ;

	} else if(intersection.size() == 1) { 

	    QPoint3f * i = intersection.first() ;

	    foreach(QPoint3f * p, intersection) { p->setColor(Qt::yellow) ; p->setSize(0.01) ; ObjectManager::instance()->addObject(p) ; }

	    QList<QPoint3f *> intersections1 = iSynaps->intersectionPoints(e1, cell) ;
	    QList<QPoint3f *> intersections2 = iSynaps->intersectionPoints(e2, cell) ;

	    // Check we are in expected situation
 	    Q_ASSERT(intersections1.size() == 2) ;
 	    Q_ASSERT(intersections2.size() == 2) ;

	    // Visualization
	    // bb->setColor(Qt::red) ;
	    // ObjectManager::instance()->addObject(bb) ;	
	    // ObjectManager::instance()->addObject(i) ;
	    // foreach(QPoint3f * p, intersections1) { p->setSize(0.01) ; ObjectManager::instance()->addObject(p) ; }
	    // foreach(QPoint3f * p, intersections2) { p->setSize(0.01) ; ObjectManager::instance()->addObject(p) ; }
	    
	    // Fill maps
	    cps.insert(consecutivePointPosition(i, e1,  this), i) ;
	    cqs.insert(consecutivePointPosition(i, e2, other), i) ;
	    // cout << FG_YELLOW << "Computing Next Region Flag for " << i << " : " << endl ;
	    nrg.insert(i, (nextRegionFlag(intersections1, intersections2, e1, e2, bb, Intersection) == e1) ? 0 : 1) ;
	    // cout << "Result : " << nrg.value(i) << NOCOLOR << endl ;
	    rvt.insert(i, new QRegionVertex(i)) ;
	    ops.insert(i, e1->object()) ;
	    oqs.insert(i, e2->object()) ;
	
	} else {
	    
	    // Visualization
// 	    n1->cell->setColor(Qt::lightGray) ;
// 	    n2->cell->setColor(Qt::lightGray) ;
// 	    bb->setColor(Qt::red) ;
// 	    ObjectManager::instance()->addObject(n1->cell) ;
// 	    ObjectManager::instance()->addObject(n2->cell) ;
// 	    ObjectManager::instance()->addObject(bb) ;	

	    // The conflict zone contains more than *one* intersection point.
	    // It has to be subdivided further, leading to smaller conflict zones (as QSegmentationQuery_t's)
	    // in which there is at most *one* intersection point.
	    query << QRegionSegmentation::refine(q) ;
	}
    }
    
# ifdef INTDEBUG
    // Check maps
    cout << "CPs: " << flush ; foreach(QPoint3f * p, cps.values()) cout << p->name().toAscii().constData() << " | " << flush ; cout << endl ;
    cout << "CPs: " << flush ; foreach(float i, cps.keys()) cout << i << " | " << flush ; cout << endl ;
    cout << "CQs: " << flush ; foreach(QPoint3f * p, cqs.values()) cout << p->name().toAscii().constData() << " | " << flush ; cout << endl ;
    cout << "CQs: " << flush ; foreach(float i, cqs.keys()) cout << i << " | " << flush ; cout << endl ;
    cout << "NRg: " << flush ; foreach(QPoint3f * p, nrg.keys()) cout << p->name().toAscii().constData() << " | " << flush ; cout << endl ;
    cout << "NRg: " << flush ; foreach(int i, nrg.values()) cout << i << " | " << flush ; cout << endl ;
    cout << endl ;
# endif

/*! Compute regions
  
Consider that we are moving from the intersection point x to the y. There are two possibilities: 
a) y is immediate right neighbour of x. In this case CP(x) < CP(y) and all polygon vertices with indexes between ceil(x) and floor(y) are added to the result. 
b) the right immediate neighbour of x is pointer to the y (see Figure 8). This means that CP(x) > CP(y). The polygon vertices between x and y are obtained in two steps:
   - vertices with indexes i: ceil(x) <= i <= last vertex index in the loop (or appropriate ring)
   - vertices with indexes j: the last vertex index in the loop(or appropriate ring) vertex index < j < y.

At the beginning we arbitrarily choose the first ipt from nrg. It is then removed from nrg.
Then, next ipt are chosen among cps | cqs and corresonding entries are removed from nrg.

*/

# ifdef INTDEBUG
    cout << "Checking if r1 is well formed ..." << flush ;
    Q_ASSERT(r1->vertices().size() == r1->edges().size()) ;
    cout << "[" << FG_GREEN << "Ok" << NOCOLOR << "]" << endl ;

    cout << "Checking if r2 is well formed ..." << flush ;
    Q_ASSERT(r2->vertices().size() == r2->edges().size()) ;
    cout << "[" << FG_GREEN << "Ok" << NOCOLOR << "]" << endl ;
# endif

    while(nrg.size()) {

	QRegion3f * f = new QRegion3f ;
	QPoint3f * f_ipt = nrg.keys().first() ;
	QPoint3f * c_ipt = 0 ; int c_table ; int c_index ; float c_value ; int c_ceil ;
	QPoint3f * n_ipt ;     int n_table ; int n_index ; float n_value ; int n_floor ;
	int firstIndexInLoop, lastIndexInLoop ;

	Q_UNUSED(n_table) ;

	forever {

	    if(!c_ipt) c_ipt = f_ipt ;

	    c_table = nrg.take(c_ipt) ;

# ifdef INTDEBUG
	    cout << "Processing " << c_ipt << endl ;
# endif

	    switch(c_table) {

	    case 0:
		
# ifdef INTDEBUG
		cout << "Table 0:" << endl ;
# endif

		c_index = cps.values().indexOf(c_ipt) ;
		c_value = cps.keys().at(c_index) ;
		c_ceil  = (int)(c_value+1) ;

# ifdef INTDEBUG	       
 		cout << "c_index: " << c_index << endl ;
 		cout << "c_value: " << c_value << endl ;
 		cout << "c_ceil: "  << c_ceil  << endl ;
# endif

		firstIndexInLoop = r1->firstIndexInLoopFrom((int)c_value) ;
		lastIndexInLoop  =  r1->lastIndexInLoopFrom((int)c_value) ;

# ifdef INTDEBUG
  		cout << "firstIndexInLoop: " << firstIndexInLoop << endl ;
  		cout << "lastIndexInLoop: " << lastIndexInLoop << endl ;
# endif

		Q_ASSERT(firstIndexInLoop >= 0) ;
		Q_ASSERT( lastIndexInLoop >= 0) ;

		if(c_index+1 < cps.size()) {
		    n_index = c_index+1 ;
		} else {
		    n_index = 0 ;
		    while(cps.keys().at(n_index) < firstIndexInLoop) n_index++ ;
		}

		Q_ASSERT(n_index < cps.size()) ;

		n_value = cps.keys().at(n_index) ;
		n_ipt = cps.values().at(n_index) ;
		n_floor = (int)(n_value) ;

# ifdef INTDEBUG
 		cout << "n_index: " << n_index << endl ;
 		cout << "n_value: " << n_value << endl ;
 		cout << "n_ipt: "   << n_ipt  << endl ;
  		cout << "n_floor: " << n_floor << endl ;
# endif

		if(n_value < c_value) {
		    
		    f->addVertex(rvt.value(c_ipt)) ;
		
		    if(c_value < lastIndexInLoop && firstIndexInLoop < n_value) {
			f->addEdge(rvt.value(c_ipt), r1->vertices().at(c_ceil), ops.value(c_ipt)) ;
			f->addEdge(r1->vertices().at(lastIndexInLoop), r1->vertices().at(firstIndexInLoop), r1->edges().at(lastIndexInLoop)->object()) ;
			f->addEdge(r1->vertices().at(n_floor), rvt.value(n_ipt), r1->edges().at(n_floor)->object()) ;
		    } else if(c_value < lastIndexInLoop) {
			f->addEdge(rvt.value(c_ipt), r1->vertices().at(c_ceil), ops.value(c_ipt)) ;
			f->addEdge(r1->vertices().at(lastIndexInLoop), rvt.value(n_ipt), r1->edges().at(lastIndexInLoop)->object()) ;
		    } else if(firstIndexInLoop < n_value) {
			f->addEdge(rvt.value(c_ipt), r1->vertices().at(firstIndexInLoop), ops.value(c_ipt)) ;
			f->addEdge(r1->vertices().at(n_floor), rvt.value(n_ipt), r1->edges().at(n_floor)->object()) ;
		    } else {
			f->addEdge(rvt.value(c_ipt), rvt.value(n_ipt), ops.value(c_ipt)) ;
		    }
		    
		    for(int i = c_ceil ; i < lastIndexInLoop ; i++) {
			f->addVertex(r1->vertices().at(i)) ;
			f->addEdge(r1->edges().at(i)) ;
		    }

		    for(int i = firstIndexInLoop ; i < n_floor ; i++) {
			f->addVertex(r1->vertices().at(i)) ;
			f->addEdge(r1->edges().at(i)) ;
		    }

		} else {

		    f->addVertex(rvt.value(c_ipt)) ;

		    if(c_ceil < n_value) {
			f->addEdge(rvt.value(c_ipt), r1->vertices().at(c_ceil), ops.value(c_ipt)) ;
			f->addEdge(r1->vertices().at(n_floor), rvt.value(n_ipt), r1->edges().at(n_floor)->object()) ;
		    } else {
			f->addEdge(rvt.value(c_ipt), rvt.value(n_ipt), ops.value(c_ipt)) ;
		    }
		    
		    for(int i = c_ceil ; i < n_floor ; i++) {
			f->addVertex(r1->vertices().at(i)) ;
			f->addEdge(r1->edges().at(i)) ;
		    }

		}

		break ;
		
	    case 1: 

# ifdef INTDEBUG	       
		cout << "Table 1:" << endl ;
# endif

		c_index = cqs.values().indexOf(c_ipt) ;
		c_value = cqs.keys().at(c_index) ;
		c_ceil  = (int)(c_value+1) ;

# ifdef INTDEBUG		
   		cout << "c_index: " << c_index << endl ;
  		cout << "c_value: " << c_value << endl ;
  		cout << "c_ceil: "  << c_ceil  << endl ;
# endif

		firstIndexInLoop = r2->firstIndexInLoopFrom((int)c_value) ;
		lastIndexInLoop  =  r2->lastIndexInLoopFrom((int)c_value) ;

# ifdef INTDEBUG
  		cout << "firstIndexInLoop: " << firstIndexInLoop << endl ;
  		cout << "lastIndexInLoop: " << lastIndexInLoop << endl ;
# endif

		Q_ASSERT(firstIndexInLoop >= 0) ;
		Q_ASSERT( lastIndexInLoop >= 0) ;

		if(c_index+1 < cqs.size()) {
		    n_index = c_index+1 ;
		} else {
		    n_index = 0 ;
		    while(cqs.keys().at(n_index) < firstIndexInLoop) n_index++ ;
		}

		Q_ASSERT(n_index < cqs.size()) ;

		n_value = cqs.keys().at(n_index) ;
		n_ipt = cqs.values().at(n_index) ;
		n_floor = (int)(n_value) ;

# ifdef INTDEBUG
  		cout << "n_index: " << n_index << endl ;
  		cout << "n_value: " << n_value << endl ;
  		cout << "n_ipt: "   << n_ipt  << endl ;
  		cout << "n_floor: " << n_floor << endl ;
# endif

		if(n_value < c_value) {
		    
		    f->addVertex(rvt.value(c_ipt)) ;
		
		    if(c_value < lastIndexInLoop && firstIndexInLoop < n_value) {
			f->addEdge(rvt.value(c_ipt), r2->vertices().at(c_ceil), oqs.value(c_ipt)) ;
			f->addEdge(r2->vertices().at(lastIndexInLoop), r2->vertices().at(firstIndexInLoop), r2->edges().at(lastIndexInLoop)->object()) ;
			f->addEdge(r2->vertices().at(n_floor), rvt.value(n_ipt), r2->edges().at(n_floor)->object()) ;
		    } else if(c_value < lastIndexInLoop) {
			f->addEdge(rvt.value(c_ipt), r2->vertices().at(c_ceil), oqs.value(c_ipt)) ;
			f->addEdge(r2->vertices().at(lastIndexInLoop), rvt.value(n_ipt), r2->edges().at(lastIndexInLoop)->object()) ;
		    } else if(firstIndexInLoop < n_value) {
			f->addEdge(rvt.value(c_ipt), r2->vertices().at(firstIndexInLoop), oqs.value(c_ipt)) ;
			f->addEdge(r2->vertices().at(n_floor), rvt.value(n_ipt), r2->edges().at(n_floor)->object()) ;
		    } else {
			f->addEdge(rvt.value(c_ipt), rvt.value(n_ipt), oqs.value(c_ipt)) ;
		    }
		    
		    for(int i = c_ceil ; i < lastIndexInLoop ; i++) {
			f->addVertex(r2->vertices().at(i)) ;
			f->addEdge(r2->edges().at(i)) ;
		    }

		    for(int i = firstIndexInLoop ; i < n_floor ; i++) {
			f->addVertex(r2->vertices().at(i)) ;
			f->addEdge(r2->edges().at(i)) ;
		    }

		} else {

		    f->addVertex(rvt.value(c_ipt)) ;

		    if(c_ceil < n_value) {
			f->addEdge(rvt.value(c_ipt), r2->vertices().at(c_ceil), oqs.value(c_ipt)) ;
			f->addEdge(r2->vertices().at(n_floor), rvt.value(n_ipt), r2->edges().at(n_floor)->object()) ;
		    } else {
			f->addEdge(rvt.value(c_ipt), rvt.value(n_ipt), oqs.value(c_ipt)) ;
		    }

		    for(int i = c_ceil ; i < n_floor ; i++) {
			f->addVertex(r2->vertices().at(i)) ;
			f->addEdge(r2->edges().at(i)) ;
		    }

		}

		break ;

	    } 
	    
	    c_ipt = n_ipt ;

# ifdef INTDEBUG	
	    cout << endl ;
# endif

	    if(n_ipt == f_ipt) break ;
	}
	
	// Set determinors and killers
	foreach(QAxelObject * o, r1->determinors()) if(!f->determinors().contains(o)) f->addDeterminor(o) ;
	foreach(QAxelObject * o, r2->determinors()) if(!f->determinors().contains(o)) f->addDeterminor(o) ;

	f->segmentation()->build() ;

	l << f ;
    }

    return l ;
}

/*************************************************************
 * Returns a distance metric relative to the index of the 
 * source vertex of the edge.
 *
 * CAUTION: This metric can fail in VERY degenerate cases.
 *************************************************************/
float QRegion3f::consecutivePointPosition(QPoint3f * intersection, QRegionEdge * e, QRegion3f * r)
{
    float norm  = (e->dest()->x() - e->source()->x()) * (e->dest()->x() - e->source()->x()) + (e->dest()->y() - e->source()->y()) * (e->dest()->y() - e->source()->y()) ;
    float value = (intersection->x() - e->source()->x()) * (intersection->x() - e->source()->x()) + (intersection->y() - e->source()->y()) * (intersection->y() - e->source()->y()) ;
# ifdef CPPDEBUG
    cout << "Checking for index of " << e->source() << " in " << r->name().toAscii().constData() << endl ;
# endif
    Q_ASSERT(r->vertices().contains(e->source())) ;
    return value/norm + r->vertices().indexOf(e->source()) ;
}

/*************************************************************
 * Returns the edge along which to continue the walkabout
 *************************************************************/
QRegionEdge * QRegion3f::nextRegionFlag(QList<QPoint3f *> l1, QList<QPoint3f *> l2, QRegionEdge * e1, QRegionEdge * e2, QBoundingBox * bb, Operation op)
{
    Q_ASSERT(l1.size() == 2) ;
    Q_ASSERT(l2.size() == 2) ;

    // Get the list of inputPoints and outputPoints
    QList<QPoint3f *> inputPoints ;
    QList<QPoint3f *> outputPoints ;
# ifdef NRGDEBUG
    cout << "e1 code : " << e1->codeString().toAscii().constData() << " : l1.size() = " << l1.size() << " " << flush ; foreach(QPoint3f * p, l1) cout << p->name().toAscii().constData() << " " << flush ; cout << endl ;
    cout << "e2 code : " << e2->codeString().toAscii().constData() << " : l2.size() = " << l2.size() << " " << flush ; foreach(QPoint3f * p, l2) cout << p->name().toAscii().constData() << " " << flush ; cout << endl ;
# endif
    if(e1->code() == QRegionEdge::PlusPlus)   { if (l1.first()->x() <  l1.last()->x() && l1.first()->y() <  l1.last()->y()) { inputPoints << l1.first() ; outputPoints << l1.last() ; } else { inputPoints << l1.last() ; outputPoints << l1.first() ; } }
    if(e1->code() == QRegionEdge::PlusMinus)  { if (l1.first()->x() <  l1.last()->x() && l1.first()->y() >  l1.last()->y()) { inputPoints << l1.first() ; outputPoints << l1.last() ; } else { inputPoints << l1.last() ; outputPoints << l1.first() ; } }
    if(e1->code() == QRegionEdge::MinusPlus)  { if (l1.first()->x() >  l1.last()->x() && l1.first()->y() <  l1.last()->y()) { inputPoints << l1.first() ; outputPoints << l1.last() ; } else { inputPoints << l1.last() ; outputPoints << l1.first() ; } }
    if(e1->code() == QRegionEdge::MinusMinus) { if (l1.first()->x() >  l1.last()->x() && l1.first()->y() >  l1.last()->y()) { inputPoints << l1.first() ; outputPoints << l1.last() ; } else { inputPoints << l1.last() ; outputPoints << l1.first() ; } }
    if(e1->code() == QRegionEdge::NullPlus)   { if (l1.first()->x() == l1.last()->x() && l1.first()->y() <  l1.last()->y()) { inputPoints << l1.first() ; outputPoints << l1.last() ; } else { inputPoints << l1.last() ; outputPoints << l1.first() ; } }
    if(e1->code() == QRegionEdge::NullMinus)  { if (l1.first()->x() == l1.last()->x() && l1.first()->y() >  l1.last()->y()) { inputPoints << l1.first() ; outputPoints << l1.last() ; } else { inputPoints << l1.last() ; outputPoints << l1.first() ; } }
    if(e1->code() == QRegionEdge::PlusNull)   { if (l1.first()->x() <  l1.last()->x() && l1.first()->y() == l1.last()->y()) { inputPoints << l1.first() ; outputPoints << l1.last() ; } else { inputPoints << l1.last() ; outputPoints << l1.first() ; } }
    if(e1->code() == QRegionEdge::MinusNull)  { if (l1.first()->x() >  l1.last()->x() && l1.first()->y() == l1.last()->y()) { inputPoints << l1.first() ; outputPoints << l1.last() ; } else { inputPoints << l1.last() ; outputPoints << l1.first() ; } }
    if(e1->code() == QRegionEdge::NullNull)   { if (l1.first()->x() == l1.last()->x() && l1.first()->y() == l1.last()->y()) { inputPoints << l1.first() ; outputPoints << l1.last() ; } else { inputPoints << l1.last() ; outputPoints << l1.first() ; } }

    if(e2->code() == QRegionEdge::PlusPlus)   { if (l2.first()->x() <  l2.last()->x() && l2.first()->y() <  l2.last()->y()) { inputPoints << l2.first() ; outputPoints << l2.last() ; } else { inputPoints << l2.last() ; outputPoints << l2.first() ; } }
    if(e2->code() == QRegionEdge::PlusMinus)  { if (l2.first()->x() <  l2.last()->x() && l2.first()->y() >  l2.last()->y()) { inputPoints << l2.first() ; outputPoints << l2.last() ; } else { inputPoints << l2.last() ; outputPoints << l2.first() ; } }
    if(e2->code() == QRegionEdge::MinusPlus)  { if (l2.first()->x() >  l2.last()->x() && l2.first()->y() <  l2.last()->y()) { inputPoints << l2.first() ; outputPoints << l2.last() ; } else { inputPoints << l2.last() ; outputPoints << l2.first() ; } }
    if(e2->code() == QRegionEdge::MinusMinus) { if (l2.first()->x() >  l2.last()->x() && l2.first()->y() >  l2.last()->y()) { inputPoints << l2.first() ; outputPoints << l2.last() ; } else { inputPoints << l2.last() ; outputPoints << l2.first() ; } }
    if(e2->code() == QRegionEdge::NullPlus)   { if (l2.first()->x() == l2.last()->x() && l2.first()->y() <  l2.last()->y()) { inputPoints << l2.first() ; outputPoints << l2.last() ; } else { inputPoints << l2.last() ; outputPoints << l2.first() ; } }
    if(e2->code() == QRegionEdge::NullMinus)  { if (l2.first()->x() == l2.last()->x() && l2.first()->y() >  l2.last()->y()) { inputPoints << l2.first() ; outputPoints << l2.last() ; } else { inputPoints << l2.last() ; outputPoints << l2.first() ; } }
    if(e2->code() == QRegionEdge::PlusNull)   { if (l2.first()->x() <  l2.last()->x() && l2.first()->y() == l2.last()->y()) { inputPoints << l2.first() ; outputPoints << l2.last() ; } else { inputPoints << l2.last() ; outputPoints << l2.first() ; } }
    if(e2->code() == QRegionEdge::MinusNull)  { if (l2.first()->x() >  l2.last()->x() && l2.first()->y() == l2.last()->y()) { inputPoints << l2.first() ; outputPoints << l2.last() ; } else { inputPoints << l2.last() ; outputPoints << l2.first() ; } }
    if(e2->code() == QRegionEdge::NullNull)   { if (l2.first()->x() == l2.last()->x() && l2.first()->y() == l2.last()->y()) { inputPoints << l2.first() ; outputPoints << l2.last() ; } else { inputPoints << l2.last() ; outputPoints << l2.first() ; } }
# ifdef NRGDEBUG
    cout << "inputPoints: " << flush ; foreach(QPoint3f * p, inputPoints)   { p->setColor(Qt::yellow) ;  cout << p->name().toAscii().constData() << " " << flush ; } cout << endl ;
    cout << "outputPoints: " << flush ; foreach(QPoint3f * p, outputPoints) { p->setColor(Qt::magenta) ; cout << p->name().toAscii().constData() << " " << flush ; } cout << endl ;
# endif
    Q_ASSERT(inputPoints.size()  == 2) ;
    Q_ASSERT(outputPoints.size() == 2) ;
    
    // Compute a CCW repartition on the border of bb, beginning with the first input point
    QList<QPoint3f *> repartition = ccwRepartition(QList<QPoint3f *>() << inputPoints << outputPoints, bb) ;
# ifdef NRGDEBUG
    cout << "repartition computed of size " << repartition.size() << endl ;
# endif
    while(outputPoints.contains(repartition.first())) repartition << repartition.takeFirst() ;
# ifdef NRGDEBUG
    cout << "CCW repartition: " << flush ; foreach(QPoint3f * p, repartition) cout << p->name().toAscii().constData() << flush ; cout << endl ;
# endif
    foreach(QPoint3f * p, inputPoints) repartition.removeAll(p) ;
# ifdef NRGDEBUG
    cout << "CCW repartition with only output points: " << flush ; foreach(QPoint3f * p, repartition) cout << p->name().toAscii().constData() << flush ; cout << endl ;
# endif

    // Switch the operation and return the edge to continue along
    QRegionEdge * next = NULL ;

    switch(op) {
    case Union:
	if(l1.contains(repartition.first())) 
	    next = e1 ;
	else
	    next = e2 ;
	break ;
    case Intersection:
	if(l1.contains(repartition.last())) 
	    next = e1 ;
	else
	    next = e2 ;
	break ;
    }
    
    return next ;
}

QPoint3f * QRegion3f::xsmallest(QList<QPoint3f *>& l) 
{
    QPoint3f * s = l[0] ; int n = 0, i = 0 ;
    foreach(QPoint3f * p, l) { if(p->x() < s->x()) { s = p ; n = i ; } i++ ; }
    return l.takeAt(n) ;
}

QPoint3f * QRegion3f::xgreatest(QList<QPoint3f *>& l) 
{
    QPoint3f * s = l[0] ; int n = 0, i = 0 ;
    foreach(QPoint3f * p, l) { if(p->x() > s->x()) { s = p ; n = i ; } i++ ; }
    return l.takeAt(n) ;
}

QPoint3f * QRegion3f::ysmallest(QList<QPoint3f *>& l) 
{
    QPoint3f * s = l[0] ; int n = 0, i = 0 ;
    foreach(QPoint3f * p, l) { if(p->y() < s->y()) { s = p ; n = i ; } i++ ; }
    return l.takeAt(n) ;
}

QPoint3f * QRegion3f::ygreatest(QList<QPoint3f *>& l) 
{
    QPoint3f * s = l[0] ; int n = 0, i = 0 ;
    foreach(QPoint3f * p, l) { if(p->y() > s->y()) { s = p ; n = i ; } i++ ; }
    return l.takeAt(n) ;
}

/*************************************************************
 *! Returns a repartition of \param points around the box
 *  \param bb.
 *************************************************************/
QList<QPoint3f *> QRegion3f::ccwRepartition(QList<QPoint3f *> & points, QBoundingBox * bb)
{
//     cout << "Computing CCW Repartition for points : " << flush ; foreach(QPoint3f * p, points) cout << p << " | " << flush ; cout << endl ;
//     cout << "in box " << bb << endl ;

    QList<QPoint3f *> northList, westList, southList, eastList ;
    foreach(QPoint3f * ipt, points) if(qAbs(ipt->y() - bb->ymax()) < 1e-5) { northList << ipt ; points.removeAll(ipt) ; }
    foreach(QPoint3f * ipt, points) if(qAbs(ipt->x() - bb->xmin()) < 1e-5) { westList  << ipt ; points.removeAll(ipt) ; }
    foreach(QPoint3f * ipt, points) if(qAbs(ipt->y() - bb->ymin()) < 1e-5) { southList << ipt ; points.removeAll(ipt) ; }
    foreach(QPoint3f * ipt, points) if(qAbs(ipt->x() - bb->xmax()) < 1e-5) { eastList  << ipt ; points.removeAll(ipt) ; }

//     cout << __func__ << " northList.size() = " << northList.size() << endl ;
//     cout << __func__ << " westList.size() = " << westList.size() << endl ;
//     cout << __func__ << " southList.size() = " << southList.size() << endl ;
//     cout << __func__ << " eastList.size() = " << eastList.size() << endl ;

    QList<QPoint3f *> repartition ;
    while(!northList.empty()) repartition << xgreatest(northList) ;
    while(!westList.empty())  repartition << ygreatest(westList)  ;
    while(!southList.empty()) repartition << xsmallest(southList) ;
    while(!eastList.empty())  repartition << ysmallest(eastList)  ;
    
    return repartition ;
}

/*************************************************************
 * 
 *************************************************************/
bool QRegion3f::intersects(QRegion3f * other)
{
    QRegionSegmentation * r_seg1 =  this->segmentation() ;
    QRegionSegmentation * r_seg2 = other->segmentation() ;

    if(!r_seg1->isValid()) r_seg1->build() ;
    if(!r_seg2->isValid()) r_seg2->build() ;

    return (!QRegionSegmentation::intersect(r_seg1, r_seg2).empty()) ;
}

QList<QRegion3f *> QRegion3f::unite(QRegion3f * other)
{
    Q_UNUSED(other) ;
    return QList<QRegion3f *>() ;
}

bool QRegion3f::unites(QRegion3f * other)
{
    Q_UNUSED(other) ;

    if(!r_segmentation) r_segmentation->build() ;

    return true ;
}

bool QRegion3f::contains(QPoint3f * p)
{
    QSynapsInterface * iSynaps = dynamic_cast<QSynapsInterface *>(PluginManager::instance()->plugin("Synaps")) ;

    if(!r_segmentation->isValid()) r_segmentation->build() ;

    QList<QSegmentationNode_t *> leaves = r_segmentation->leaves() ;

    QRegionEdge * halfLine = new QRegionEdge(new QRegionVertex(p->x(), p->y()), new QRegionVertex(rightMostVertex()->x()+100, p->y()), new QCell) ;

    int nbi = 0 ;
    foreach(QSegmentationNode_t * n, leaves)
	if(n->cell->xmin() > p->x() && n->cell->ymin() < p->y() && n->cell->ymax() > p->y())
	    nbi += iSynaps->intersection(halfLine, n->object(), n->cell).size() ;

    return (nbi > 0 && nbi%2 == 1) ;
}

bool QRegion3f::opened(void)
{
    foreach(QRegionEdge * e, r_edges)
	if(e->type() == "Cell")
	    return true ;

    return false ;
}

bool QRegion3f::closed(void)
{
    foreach(QRegionEdge * e, r_edges)
	if(e->type() == "Cell")
	    return false ;

    return true ;
}

} ;
