/*****************************************************************************
 * A x e l
 *****************************************************************************
 * Viewer.h
 * 2006-03-31
 * Julien Wintz
 *****************************************************************************
 *               Copyright (C) 2006 INRIA Sophia-Antipolis
 *****************************************************************************
 * Comments :
 ****************************************************************************/

# ifndef VIEWER_H
# define VIEWER_H

# include <QBrush>
# include <QFont>
# include <QImage>
# include <QPen>
# include <QImage>
# include <QGLViewer/qglviewer.h>
# include <QAxel/QAxelObject.h>
# include <QAxel/QPoint3f.h>
# include <QAxel/ObjectManager.h>
# include <QGui/ToolPropertiesWidgets.h>

using namespace QAxel;
using namespace qglviewer;

class Viewer : public QGLViewer 
{
    Q_OBJECT
    Q_PROPERTY(MeshEditionMode meshEditionMode READ meshEditionMode WRITE setMeshEditionMode)
    Q_ENUMS(MeshEditionMode)
    Q_OVERRIDE(QString objectName SCRIPTABLE true)
	
public:
    Viewer(QWidget *parent);

public:
    virtual void init();
    virtual void draw();
    virtual void animate() ;
    virtual QString helpString() const;
    
    // Selection functions
    virtual void drawWithNames();
    virtual void endSelection(const QPoint&);
    
    /*! @name Grid drawing */ 
    void drawXYGrid(float size=1.0f, int nbSubdivisions=10) ;
    void drawXZGrid(float size=1.0f, int nbSubdivisions=10) ;
    void drawYZGrid(float size=1.0f, int nbSubdivisions=10) ;
    
    /*! @name Events */
    virtual bool event(QEvent * e) ;
    virtual void resizeGL(int width,int  height) ;
    virtual void keyPressEvent(QKeyEvent *e) ;
    virtual void mousePressEvent(QMouseEvent *e) ;
    virtual void mouseMoveEvent(QMouseEvent *e) ;
    virtual void mouseReleaseEvent(QMouseEvent *e) ;
    virtual void mouseDoubleClickEvent(QMouseEvent *e) ;
    
	
signals:
    /*! @name Signals */
    void mouseMoved(QPoint3f *) ;
    void prout(qglviewer::Vec translation) ;
    void objectAdded(QAxelObject *) ;
    void objectRemoved(void) ;
    void sceneChanged(void);
    void noObjectSelected(void) ;
    void objectSelected(QAxelObject *) ;
    void manipulation(QAxelObject *) ;
    void selection(int n) ;
    void pointClicked(const QPoint &) ;
    void cameraModeChanged(QString) ;
    void dimensionModeChanged(QString) ;

    //************** AF SandBox
    void directObjectAdded(const Vec &v);
    //*************************
    
public slots :
    /*! @name Drawing slots */
    void setXYGridIsDrawn(bool b) ;
    void setXZGridIsDrawn(bool b) ;
    void setYZGridIsDrawn(bool b) ;
    void setWireDisplay(void) ;
    void setLitWireDisplay(void) ;
    void setFlatDisplay(void) ;
    void setSmoothDisplay(void) ;
    void toggleMeshEditionMode(void) ;
    void removeSelectedObjects(void) ;
    void removeTemporaryObjects(void) ;
    void selectAll(void) ;
    void unselectAll(void) ;
    
    void set2dMode(void) ;
    void set3dMode(void) ;

    // Object management functions
    void addObject(QAxelObject *object);
    void addObjects(QList<QAxelObject *> objectList);
    void removeObject(QAxelObject *object);
    void updateSelection(void);
    void showObject(QAxelObject * object, int msec) ;
    
    void addObjectToSelection(QAxelObject *object);
    
    QPoint3f * convertClickToPoint(QMouseEvent * e) ;

    QAxelObject * getObjectFromName(QString name);
    
    QList<QAxelObject *> listObjects(void) ;
    QList<QAxelObject *> listSelectedObjects(void) ;
    QList<QAxelObject *> listGrabbedObjects(void) ;
    QList<QAxelObject *> listExtrusionObjects(void) ;
    QList<QAxelObject *> listObjectsFromType(QString type) ;
    QList<QAxelObject *> listSelectedObjectsFromType(QString type) ;
    QList<QAxelObject *> listEditedObjectsFromType(QString type) ;
    QList<QAxelObject *> listExtrusionObjectsFromType(QString type) ;
    QList<QAxelObject *> listGrabbedObjectsFromType(QString type) ;
    
    //************** AF SandBox
    void selectFromName(QString objectName) ;
    void setSelectedObjects(QList<QAxelObject *> select);
    QList<QString> getListObjectsName(void);
  //*************************/

    void setOrthographicDisplay(void) ;
    void setPerspectiveDisplay(void) ;
    void setCameraPosition(float x, float y, float z) ;
    void setCameraUpVector(float x, float y, float z) ;
    void setCameraViewDirection(float x, float y, float z) ;
    void setCameraConstraint(CameraConstraint * constraint) ;
    
public slots:
    void refineImplicitCurves(void) ;
    void intersectRationalSurfaces(void) ;
    void intersectParametricSurfaces(void) ;
    void intersectImplicitCurves(void) ;
    void intersectImplicitSurfaces(void) ;
    void autoIntersectParametricSurfaces(void) ;
    void interpolatePointsAsBSpline(void) ;
    void approximatePointsAsBSplineCurve(void) ;
    void approximatePointsAsBSplineSurface(int, int, int, int, int) ;
    void joinBSplineCurves(void) ;
    void intersectBSplineCurves(void) ;
    void offsetBSplineCurves(void) ;
    void blendBSplineCurves(void) ;
    void quadClipImplicitCurves(void) ;


protected:
    void dropEvent(QDropEvent * event) ;
    void dragEnterEvent(QDragEnterEvent * event) ;

private :
    void startManipulation(void) ;
    void drawSelectionRectangle(void) const ;
    void removeObjectFromSelection(QAxelObject *object);
    void clearSelection(void) ;

public :
    enum MeshEditionMode { VertexEdition, EdgeEdition, FaceEdition } ;   
    enum FaceCullingMode { Front = GL_FRONT, Back  = GL_BACK, FrontAndBack = GL_FRONT_AND_BACK } ;
    void setMeshEditionMode(MeshEditionMode mode) ;
    void setFaceCulling(bool enable) ;
    void setDepthTest(bool enable) ;
    void setSmoothShading(bool enable) ;
    void setAntiAliasing(bool enable) ;
    MeshEditionMode meshEditionMode() const { return meshEditionMode_ ; }
    FaceCullingMode faceCullingMode() const { return faceCullingMode_ ; }
    
private :
    QRect rectangle_ ;

    enum SelectionMode { NONE, ADD, REMOVE, DIRECT } ;
    SelectionMode selectionMode_ ;
    
    QList<QAxelObject*> selection_;
    QList<QAxelObject*> objects_;
    QList<QAxelObject*> temporary_;
    
    QMap<QString, QAxelObject *> nameMap_;

    MeshEditionMode meshEditionMode_ ;
    FaceCullingMode faceCullingMode_ ;
    
    bool xyGridIsDrawn_ ;
    bool xzGridIsDrawn_ ;
    bool yzGridIsDrawn_ ;
    bool wireframe_, flatShading_ ;
    bool addObjectDirectly ;
    bool directMode ;
	
private:
    CameraConstraint * cameraConstraint ;

private:    
    QAction * refineImplicitCurveAction ;
    QAction * autoIntersectParametricSurfacesAction ;
    QAction * intersectRationalSurfacesAction ;
    QAction * intersectParametricSurfacesAction ;
    QAction * intersectImplicitCurvesAction ;
    QAction * intersectImplicitSurfacesAction ;
    QAction * interpolatePointsAsBSplineAction ;
    QAction * intersectBSplineCurvesAction ;
    QAction * offsetBSplineCurvesAction ;
    QAction * blendBSplineCurvesAction ;
    QAction * quadClipImplicitCurvesAction ;
    QAction * approximatePointsAsBSplineAction ;

private:
    void formatInstructions(int width, int height);

private:
    QImage image ;

private:
    bool mf_move ;
    QPoint3f * v_oldMousePosition ;
} ;

# endif
