/*****************************************************************************
 * A x e l
 *****************************************************************************
 * ViewerSelectionEvents.h
 * 2006-11-08
 * Julien Wintz
 *****************************************************************************
 *               Copyright (C) 2006 INRIA Sophia-Antipolis
 *****************************************************************************
 * Comments :
 ****************************************************************************/

# include <QFile>
# include <QCursor>
# include <QAction>
# include <QMenu>

# include <QGui/Viewer.h>

# include <QAxel/QVertex.h>
# include <QAxel/QEdge.h>
# include <QAxel/QFace.h>
# include <QAxel/QMesh.h>

# include <QAxel/QBoundingBox.h>
# include <QAxel/QPoint3f.h>
# include <QAxel/QPlane.h>
# include <QAxel/QCurve.h>
# include <QAxel/QSurface.h>

# include <QGui/ManipulatedFrameSetConstraint.h>

void Viewer::endSelection(const QPoint& p)
{
    // Flush GL buffers
    glFlush();
    
    // Get the number of objects that were seen through the pick matrix frustum. Reset GL_RENDER mode.
    GLint nbHits = glRenderMode(GL_RENDER);
    emit selection(nbHits) ;
    
    //************** AF SandBox
    if(directMode)
    {
	if(nbHits==0)
	{
	    if(selection_.empty())
	    {
		if((nbHits==0)&&(addObjectDirectly)&&(rectangle_.width()<2))
		{
		    Vec src;
		    src.setValue(p.x(),p.y(),0.8);
		    Vec v=camera()->unprojectedCoordinatesOf(src);
		    emit directObjectAdded(v);
		}
		setMouseBinding(Qt::LeftButton, CAMERA, ROTATE);
		setMouseBinding(Qt::RightButton, CAMERA, TRANSLATE);
		setWheelBinding(Qt::NoModifier, CAMERA, ZOOM);
	    }
	    else
	    {
		unselectAll();
		emit noObjectSelected();
		setMouseBinding(Qt::LeftButton, CAMERA, ROTATE);
		setMouseBinding(Qt::RightButton, CAMERA, TRANSLATE);
		setWheelBinding(Qt::NoModifier, CAMERA, ZOOM);
	    }
	}
	else
	{
	    for (int i=0; i<nbHits; ++i)
	    {
		QAxel::QAxelObject* object = (QAxel::QAxelObject*)(selectBuffer())[4*i+3];
		switch (selectionMode_)
		{
		case DIRECT :
		case ADD :
		    addObjectToSelection(object);
		    emit objectSelected(object);
		    break;
		case REMOVE : 
		    removeObjectFromSelection(object);  
		    break;
		default : 
		    break;
		}
	    }
// 			startManipulation();
	    setMouseBinding(Qt::LeftButton, FRAME, ROTATE);
	    setMouseBinding(Qt::RightButton, FRAME, TRANSLATE);
	    setWheelBinding(Qt::NoModifier, FRAME, ZOOM);
	}
	selectionMode_ = NONE;
    }
    else
    {
	//***********************AF
	
	// Perform selection stuff over objects
	if(nbHits==0)
	{
	    unselectAll();
	    emit noObjectSelected();
	}
	else
	{
	    // Interpret results : each object created 4 values in the selectBuffer().
	    // (selectBuffer())[4*i+3] is the id pushed on the stack.
	    for (int i=0; i<nbHits; ++i)
	    {
		QAxel::QAxelObject* object = (QAxel::QAxelObject*)(selectBuffer())[4*i+3];
		switch (selectionMode_)
		{
		case ADD :
		    addObjectToSelection(object);
		    emit objectSelected(object);
		    break;
		case REMOVE : 
		    removeObjectFromSelection(object);  
		    break;
		default : 
		    break;
		}
	    }
	}
	selectionMode_ = NONE;
    }
    updateGL() ;
    
    emit pointClicked(p) ;

}

void Viewer::startManipulation(void) 
{
    Vec averagePosition;
    
    ManipulatedFrameSetConstraint* mfsc = (ManipulatedFrameSetConstraint*)(manipulatedFrame()->constraint());
    mfsc->clearSet();                   
    
    for (QList<QAxel::QAxelObject *>::const_iterator it=selection_.begin(), end=selection_.end(); it != end; ++it) 
    {
	if((*it)->type() == "Mesh") 
	    if(((QMesh *)*it)->mode() == QMesh::Edit)
		continue ;
	
	if((*it)->type() == "BSplineCurve") 
	    // Put its control points in the manipulated set
	    continue ;

	if((*it)->type() == "BSplineSurface")
	    // Put its control points in the manipulated set
	    continue ;
	
	mfsc->addObjectToSet(*it);
	averagePosition += (*it)->frame()->position();
    }
    
    if (selection_.size() > 0) 
	manipulatedFrame()->setPosition(averagePosition / selection_.size());
}

///////////////////////////////////////////////////////////////////////////////////////
//   S e l e c t i o n   t o o l s
///////////////////////////////////////////////////////////////////////////////////////

void Viewer::clearSelection(void) 
{
    for(int i=0 ; i<selection_.size() ; i++) 
    {
	selection_[i]->deselect() ;
    }
    selection_.clear() ;
}

void Viewer::addObjectToSelection(QAxel::QAxelObject *object)
{	
    if (!selection_.contains(object)) 
	selection_.push_back(object);
    object->select();
}

void Viewer::removeObjectFromSelection(QAxel::QAxelObject *object)
{
    selection_.removeAll(object);
    object->deselect() ;
}

void Viewer::drawSelectionRectangle() const 
{
    startScreenCoordinatesSystem();
    glDisable(GL_LIGHTING);

    glColor4f(0.2, 0.2, 0.2, 0.2);
    glBegin(GL_QUADS);
    glVertex2i(rectangle_.left(),  rectangle_.top());
    glVertex2i(rectangle_.right(), rectangle_.top());
    glVertex2i(rectangle_.right(), rectangle_.bottom());
    glVertex2i(rectangle_.left(),  rectangle_.bottom());
    glEnd();

    glLineWidth(2.0);
    glColor4f(0.4, 0.4, 0.5, 0.5);
    glBegin(GL_LINE_LOOP);
    glVertex2i(rectangle_.left(),  rectangle_.top());
    glVertex2i(rectangle_.right(), rectangle_.top());
    glVertex2i(rectangle_.right(), rectangle_.bottom());
    glVertex2i(rectangle_.left(),  rectangle_.bottom());
    glEnd();

    glLineWidth(1.0) ;

    glEnable(GL_LIGHTING);
    stopScreenCoordinatesSystem();
}

void Viewer::updateSelection(void) 
{
    selection_.clear() ;
    for(int i=0 ; i < objects_.size() ; i++) 
    {
	if(objects_[i]->selected()) 
	    selection_ << objects_[i];
    }
}

